"""
Interpolation of the SSH AVISO
------------------------------
"""
from . import detail
import os
import re
import datetime
import numpy as np
import pyinterp.backends.xarray
import xarray as xr
from scipy import interpolate


#LOGGER = logging.getLogger(__name__)


#@nb.njit('(float32[::1])(int64[::1], float32[:, ::1], int64[::1])', cache=True)
def _time_interp(xp: np.ndarray, yp: np.ndarray, xi: np.ndarray) -> np.ndarray:
    """Time interpolation for the different spatial grids interpolated on the
    SWOT data"""
    xp_diff = np.diff(xp)

    assert xp.shape[0] == yp.shape[0] and yp.shape[1] == xi.shape[0]
    assert np.all(xp_diff == xp_diff[0])

    result = np.empty(yp.shape[1:], dtype=yp.dtype)

    step = 1.0 / xp_diff[0]
    size = xp.size

    for ix in range(yp.shape[0]):
        index = int(np.around((xi[ix] - xp[0]) * step))
        assert index >= 0 or index <= size
        if index == size - 1:
            i0 = index - 1
            i1 = index
        else:
            i0 = index
            i1 = index + 1
        t0 = xp[i0]
        t1 = xp[i1]
        dt = t1 - t0
        w0 = (t1 - xi[ix]) / dt
        w1 = (xi[ix] - t0) / dt

        for jx in range(yp.shape[1]):
            result[jx] = (w0 * yp[i0, jx] + w1 * yp[i1, jx]) / (w0 + w1)

    return result


def _spatial_interp(z_model: xr.DataArray, x_model: xr.DataArray, y_model: xr.DataArray,
                    x_sat: np.ndarray, y_sat: np.ndarray):
    mesh = pyinterp.RTree(dtype="float32")
    x, y, z = (), (), ()

    for face in range(13):
        x_face = x_model[face, :].compute()
        y_face = y_model[face, :].compute()

        # We test if the face covers the satellite positions.
        ix0, ix1 = x_face.min(), x_face.max()
        iy0, iy1 = y_face.min(), y_face.max()

        box = pyinterp.geodetic.Box2D(pyinterp.geodetic.Point2D(ix0, iy0),
                                      pyinterp.geodetic.Point2D(ix1, iy1))
        mask = box.covered_by(x_sat, y_sat)
        if not np.any(mask == 1):
            continue

        # The undefined values are filtered
        z_face = z_model[face, :].compute()
        defined = ~np.isnan(z_face)
        x += (x_face[defined].flatten(), )
        y += (y_face[defined].flatten(), )
        z += (z_face[defined].flatten(), )

    # The tree is built and the interpolation is calculated
    mesh.packing(
        np.vstack((np.concatenate(x), np.concatenate(y))).T, np.concatenate(z))
    del x, y, z
    z, _ = mesh.inverse_distance_weighting(np.vstack((x_sat, y_sat)).T,
                                           within=True,
                                           k=11,
                                           radius=55000,
                                           num_threads=1)
    return z


class OREGON_MITGCM(detail.CartesianGridHandler):
    """
    Interpolation of the SSH AVISO OREGON MIT GCM toy simulation.
    """

    #: Decode the product date encoded from the file name.
    PATTERN = re.compile(r"OREGON_(\d{4}).nc").search

    def __init__(self, path, grid_path):
        super().__init__(path)
        self.grid_path = grid_path

    def load_ts(self):
        """Loading in memory the time axis of the time series"""
        items = []
        length = -1
        idays = 0
        for root, _, files in os.walk(self.path):
            for item in files:
                match = self.PATTERN(item)
                if match is not None:
                    date0 = datetime.datetime(2020, 3, 9)
                    idays += 1
                    datei = date0 + datetime.timedelta(days=idays)
                    filename = os.path.join(root, item)
                    items.append((datei, filename))
                    length = max(length, len(filename))

        # The time series is encoded in a structured Numpy array containing
        # the date and path to the file.
        ts = np.array(items,
                      dtype={
                          'names': ('date', 'path'),
                          'formats': ('datetime64[s]', f'U{length}')
                      })
        self.ts = ts[np.argsort(ts["date"])]

        # The frequency between the grids must be constant.
        #Not necessary as the date is manually set
        frequency = set(np.diff(self.ts["date"].astype(np.int64)))
        if len(frequency) != 1:
            raise RuntimeError(
                "Time series does not have a constant step between two "
                f"grids: {frequency} seconds")

        # The frequency is stored in order to load the grids required to
        # interpolate the SSH.
        self.dt = np.timedelta64(frequency.pop(), 's')

    def load_dataset(self, first_date: np.datetime64, last_date: np.datetime64
                     ) -> pyinterp.backends.xarray.Grid3D:
        """Loads the 3D cube describing the SSH in time and space."""
        if first_date < self.ts["date"][0] or last_date > self.ts["date"][-1]:
            raise IndexError(
                f"period [{first_date}, {last_date}] is out of range: "
                f"[{self.ts['date'][0]}, {self.ts['date'][-1]}]")
        first_date -= self.dt
        last_date += self.dt

        selected = self.ts["path"][(self.ts["date"] >= first_date)
                                   & (self.ts["date"] < last_date)]

        ds = xr.open_mfdataset(selected, concat_dim="time", combine="nested")
        dsc = xr.open_dataset(self.grid_path)
        lon = dsc.lon_rho
        lat = dsc.lat_rho
        ds.coords['lat'] = lat
        ds.coords['lon'] = lon
        ds = ds.rename({'y':'eta_rho', 'x':'xi_rho'})
        print(ds, )
        self.lon = lon
        self.lat = lat
        self.ssh = ds.H
        return None #pyinterp.backends.xarray.Grid3D(ds.H)
    
    def _grid_date(self, date: np.datetime64, shift: int):
        """Calculates the grid date immediately before or after the date
        provided"""
        if date.astype("int64") % self.dt.astype("int64") != 0:
            return date + self.dt * shift
        return date

    def interpolate(self, lon: np.ndarray, lat: np.ndarray,
                    time: np.ndarray) -> np.ndarray:
        """Interpolate the SSH to the required coordinates"""
        """Interpolate the SSH for the given coordinates"""
        self.load_ts()
        self.ts=self.ts['date']
        print(self.ts[0])
        first_date = self._grid_date(time[0], -1)
        last_date = self._grid_date(time[-1], 1)
        self.load_dataset(first_date, last_date)

        print(self.ts[0])
        if first_date < self.ts[0] or last_date > self.ts[-1]:
            raise IndexError(
                f"period [{first_date}, {last_date}] is out of range: "
                f"[{self.ts[0]}, {self.ts[-1]}]")

        # Mask for selecting data covering the time period provided.
        mask = (self.ts >= first_date) & (self.ts <= last_date)

        frame = self.ssh[mask]

        # Spatial interpolation of the SSH on the different selected grids.
        x_sat = lon
        y_sat = lat
        x_model = self.lon
        y_model = self.lat
        for index in range(len(frame)):
            z_model = client.scatter(frame[index, :])
            spatial_interp = _spatial_interp(z_model, x_model, y_model, x_sat,
                                             y_sat)

        # Time interpolation of the SSH.
        return _time_interp(self.ts[mask].astype("int64"),
                            np.stack(spatial_interp),
                            time.astype("datetime64[us]").astype("int64"))

